@@ -4,14 +4,14 @@
  * Author: Philipp Maier
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * it under the terms of the GNU Affero General Public License as published by
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  */
 
 #include <errno.h>
@@ -26,6 +26,8 @@
 #include <osmocom/core/logging.h>
 #include <osmocom/core/sockaddr_str.h>
 
+#include <osmocom/netif/stream.h>
+
 #include <osmocom/ranap/ranap_common.h>
 #include <osmocom/ranap/ranap_common_cn.h>
 #include <osmocom/ranap/ranap_common_ran.h>
@@ -107,6 +109,11 @@ struct mgw_fsm_priv {
 	ranap_message *ranap_rab_ass_req_message;
 	ranap_message *ranap_rab_ass_resp_message;
 	struct msgb *ranap_rab_ass_resp_msgb;
+	/* IP address contained in ranap_rab_ass_resp_msgb/message: */
+	struct osmo_sockaddr hnb_rtp_addr;
+	/* Number of MDCX transmitted. Used to detect current mgw conn_mode and
+	 * detect modify infinite loops: */
+	unsigned int mdcx_tx_cnt;
 
 	/* MGW context */
 	struct mgcp_client *mgcpc;
@@ -132,6 +139,7 @@ static void mgw_fsm_crcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
+	struct hnb_context *hnb_ctx = map->hnb_ctx;
 	struct osmo_sockaddr addr;
 	struct osmo_sockaddr_str addr_str;
 	RANAP_RAB_AssignmentRequestIEs_t *ies;
@@ -168,6 +176,19 @@ static void mgw_fsm_crcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	mgw_info.codecs[0] = CODEC_IUFP;
 	mgw_info.codecs_len = 1;
 
+	/* The HNB IuUP IP address & port is not yet known here (Rx RAB-Ass-Req Tx CRCX (RAN) time):
+	 * Assume and announce "remote IuUP IP address" == "remote Iuh signalling IP address" to the MGW
+	 * here, so that it can most probably select a proper IuUP local IP address to be used from the
+	 * start. In the event we receive a "remote IuUP IP address" != "remote Iuh signalling IP address"
+	 * later on during RAB-Ass-Resp, we'll update IP addr at the MGW through MDCX and if MGW decides
+	 * to use another IuUP local IP address as a result, it will be updated at the HNB through
+	 * RAB-Modify-Req. */
+	if (hnb_ctx && hnb_ctx->conn &&
+	    (rc = osmo_stream_srv_get_fd(hnb_ctx->conn)) >= 0) {
+		if (osmo_sock_get_remote_ip(rc, mgw_info.addr, sizeof(mgw_info.addr)) < 0)
+			LOGPFSML(fi, LOGL_ERROR, "Invalid Iuh IP Address\n");
+	}
+
 	mgw_fsm_priv->mgcpc = mgcp_client_pool_get(g_hnbgw->mgw_pool);
 	if (!mgw_fsm_priv->mgcpc) {
 		LOGPFSML(fi, LOGL_ERROR,
@@ -189,8 +210,6 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	const struct mgcp_conn_peer *mgw_info;
 	struct osmo_sockaddr_str addr_str;
-	struct osmo_sockaddr *addr = &mgw_fsm_priv->ci_hnb_crcx_ack_addr;
-	RANAP_RAB_AssignmentRequestIEs_t *ies;
 	int rc;
 
 	switch (event) {
@@ -208,7 +227,7 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			addr_str.af = AF_INET6;
 		addr_str.port = mgw_info->port;
 		osmo_strlcpy(addr_str.ip, mgw_info->addr, sizeof(addr_str.ip));
-		rc = osmo_sockaddr_str_to_sockaddr(&addr_str, &addr->u.sas);
+		rc = osmo_sockaddr_str_to_sockaddr(&addr_str, &mgw_fsm_priv->ci_hnb_crcx_ack_addr.u.sas);
 		if (rc < 0) {
 			LOGPFSML(fi, LOGL_ERROR,
 				 "Failed to convert RTP IP-address (%s) and Port (%u) to its binary representation\n",
@@ -217,16 +236,6 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			return;
 		}
 
-		ies = &mgw_fsm_priv->ranap_rab_ass_req_message->msg.raB_AssignmentRequestIEs;
-		rc = ranap_rab_ass_req_ies_replace_inet_addr(ies, addr, mgw_fsm_priv->rab_id);
-		if (rc < 0) {
-			LOGPFSML(fi, LOGL_ERROR,
-				 "Failed to replace RTP IP-address (%s) and Port (%u) in RAB-AssignmentRequest\n",
-				 mgw_info->addr, mgw_info->port);
-			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
-			return;
-		}
-
 		mgw_fsm_state_chg(fi, MGW_ST_ASSIGN);
 		return;
 	default:
@@ -240,8 +249,16 @@ static void mgw_fsm_assign_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
 	RANAP_RAB_AssignmentRequestIEs_t *ies;
 	struct msgb *msg;
+	int rc;
 
 	ies = &mgw_fsm_priv->ranap_rab_ass_req_message->msg.raB_AssignmentRequestIEs;
+	rc = ranap_rab_ass_req_ies_replace_inet_addr(ies, &mgw_fsm_priv->ci_hnb_crcx_ack_addr, mgw_fsm_priv->rab_id);
+	if (rc < 0) {
+		LOGPFSML(fi, LOGL_ERROR, "Failed to replace RTP IP-address and Port in RAB-AssignmentRequest\n");
+		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+		return;
+	}
+
 	msg = ranap_rab_ass_req_encode(ies);
 	if (!msg) {
 		LOGPFSML(fi, LOGL_ERROR, "failed to re-encode RAB-AssignmentRequest message\n");
@@ -257,9 +274,74 @@ static void mgw_fsm_assign_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state
 
 static void mgw_fsm_assign(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
+	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
+	struct hnbgw_context_map *map = mgw_fsm_priv->map;
+	RANAP_RAB_AssignmentResponseIEs_t *ies;
+	bool rab_failed_at_hnb;
+	struct osmo_sockaddr addr;
+	enum mgw_fsm_state next_st;
+	int rc;
+
 	switch (event) {
 	case MGW_EV_RAB_ASS_RESP:
-		mgw_fsm_state_chg(fi, MGW_ST_MDCX_HNB);
+		LOGPFSML(fi, LOGL_DEBUG, "RAB-AssignmentResponse received, completing HNB side call-leg on MGW...\n");
+		ies = &mgw_fsm_priv->ranap_rab_ass_resp_message->msg.raB_AssignmentResponseIEs;
+		rc = ranap_rab_ass_resp_ies_extract_inet_addr(&addr, ies, mgw_fsm_priv->rab_id);
+		if (rc < 0) {
+			rab_failed_at_hnb = ranap_rab_ass_resp_ies_check_failure(ies, mgw_fsm_priv->rab_id);
+			if (rab_failed_at_hnb) {
+				struct msgb *msg;
+
+				LOGPFSML(fi, LOGL_ERROR,
+					"The RAB-AssignmentResponse contains a RAB-FailedList, RAB-Assignment (%u) failed.\n",
+					mgw_fsm_priv->rab_id);
+
+				/* Forward the RAB-AssignmentResponse transparently. This will ensure that the MSC
+				 * is informed about the problem. */
+				LOGPFSML(fi, LOGL_DEBUG, "forwarding unmodified RAB-AssignmentResponse to MSC\n");
+
+				msg = mgw_fsm_priv->ranap_rab_ass_resp_msgb;
+				mgw_fsm_priv->ranap_rab_ass_resp_msgb = NULL;
+				talloc_steal(OTC_SELECT, msg);
+
+				rc = map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
+				if (rc < 0) {
+					LOGPFSML(fi, LOGL_DEBUG, "failed to forward RAB-AssignmentResponse message\n");
+					osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+				}
+
+				/* Even though this is a failure situation, we still release normally as the error is located
+				* at the HNB. */
+				osmo_fsm_inst_state_chg(fi, MGW_ST_RELEASE, 0, 0);
+				return;
+			}
+
+			/* The RAB-ID we are dealing with is not on an FailedList and we were unable to parse the response
+			* normally. This is a situation we cannot recover from. */
+			LOGPFSML(fi, LOGL_ERROR, "Failed to extract RTP IP-address and Port from RAB-AssignmentResponse\n");
+			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+			return;
+		}
+
+		/* Break infinite loops modifications between HNB and our MGW: */
+		if (mgw_fsm_priv->mdcx_tx_cnt > 3) {
+			LOGPFSML(fi, LOGL_ERROR, "IuUP addr modification infinite loop detected between HNB and MGW, "
+				 "check your network routing config -- abort!\n");
+			osmo_fsm_inst_state_chg(fi, MGW_ST_RELEASE, 0, 0);
+			return;
+		}
+
+		/* Send at least 1 MDCX in order to change conn_mode to SEND_RECV.
+		 * From there on, MDCX is only needed if HNB IP/Port changed: */
+		if (mgw_fsm_priv->mdcx_tx_cnt == 0 ||
+		    osmo_sockaddr_cmp(&addr, &mgw_fsm_priv->hnb_rtp_addr) != 0) {
+			next_st = MGW_ST_MDCX_HNB;
+		} else {
+			LOGPFSML(fi, LOGL_DEBUG, "RAB-AssignmentResponse received with unchanged IuUP attributes, skipping MDCX...\n");
+			next_st = MGW_ST_CRCX_MSC;
+		}
+		mgw_fsm_priv->hnb_rtp_addr = addr;
+		mgw_fsm_state_chg(fi, next_st);
 		return;
 	default:
 		OSMO_ASSERT(false);
@@ -271,13 +353,8 @@ static void mgw_fsm_mdcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
 	struct mgcp_conn_peer mgw_info;
-	struct osmo_sockaddr addr;
 	struct osmo_sockaddr_str addr_str;
-	RANAP_RAB_AssignmentResponseIEs_t *ies;
 	int rc;
-	bool rab_failed_at_hnb;
-
-	LOGPFSML(fi, LOGL_DEBUG, "RAB-AssignmentResponse received, completing HNB side call-leg on MGW...\n");
 
 	mgw_info = (struct mgcp_conn_peer) {
 		.call_id = map->rua_ctx_id,
@@ -287,53 +364,17 @@ static void mgw_fsm_mdcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	mgw_info.codecs[0] = CODEC_IUFP;
 	mgw_info.codecs_len = 1;
 
-	ies = &mgw_fsm_priv->ranap_rab_ass_resp_message->msg.raB_AssignmentResponseIEs;
-	rc = ranap_rab_ass_resp_ies_extract_inet_addr(&addr, ies, mgw_fsm_priv->rab_id);
-	if (rc < 0) {
-		rab_failed_at_hnb = ranap_rab_ass_resp_ies_check_failure(ies, mgw_fsm_priv->rab_id);
-		if (rab_failed_at_hnb) {
-			struct msgb *msg;
-
-			LOGPFSML(fi, LOGL_ERROR,
-				 "The RAB-AssignmentResponse contains a RAB-FailedList, RAB-Assignment (%u) failed.\n",
-				 mgw_fsm_priv->rab_id);
-
-			/* Forward the RAB-AssignmentResponse transparently. This will ensure that the MSC is informed
-			 * about the problem. */
-			LOGPFSML(fi, LOGL_DEBUG, "forwarding unmodified RAB-AssignmentResponse to MSC\n");
-
-			msg = mgw_fsm_priv->ranap_rab_ass_resp_msgb;
-			mgw_fsm_priv->ranap_rab_ass_resp_msgb = NULL;
-			talloc_steal(OTC_SELECT, msg);
-
-			rc = map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
-			if (rc < 0) {
-				LOGPFSML(fi, LOGL_DEBUG, "failed to forward RAB-AssignmentResponse message\n");
-				osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
-			}
-
-			/* Even though this is a failure situation, we still release normally as the error is located
-			 * at the HNB. */
-			osmo_fsm_inst_state_chg(fi, MGW_ST_RELEASE, 0, 0);
-			return;
-		}
-
-		/* The RAB-ID we are dealing with is not on an FailedList and we were unable to parse the response
-		 * normally. This is a situation we cannot recover from. */
-		LOGPFSML(fi, LOGL_ERROR, "Failed to extract RTP IP-address and Port from RAB-AssignmentResponse\n");
-		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
-		return;
-	}
-
-	rc = osmo_sockaddr_str_from_sockaddr(&addr_str, &addr.u.sas);
+	rc = osmo_sockaddr_str_from_sockaddr(&addr_str, &mgw_fsm_priv->hnb_rtp_addr.u.sas);
 	if (rc < 0) {
 		LOGPFSML(fi, LOGL_ERROR, "Invalid RTP IP-address or Port in RAB-AssignmentResponse\n");
 		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
 		return;
 	}
+
 	osmo_strlcpy(mgw_info.addr, addr_str.ip, sizeof(mgw_info.addr));
 	mgw_info.port = addr_str.port;
 
+	mgw_fsm_priv->mdcx_tx_cnt++;
 	osmo_mgcpc_ep_ci_request(mgw_fsm_priv->mgcpc_ep_ci_hnb, MGCP_VERB_MDCX, &mgw_info, fi, MGW_EV_MGCP_OK,
 				 MGW_EV_MGCP_FAIL, NULL);
 }
@@ -371,13 +412,14 @@ static void mgw_fsm_mdcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		}
 
 		if (osmo_sockaddr_cmp(&mgw_fsm_priv->ci_hnb_crcx_ack_addr, &addr) != 0) {
-			/* FIXME: Send RAB Modify Req to HNB. See OS#6127 */
 			char addr_buf[INET6_ADDRSTRLEN + 8];
-			LOGPFSML(fi, LOGL_ERROR, "Local MGW IuUP IP address %s changed to %s during MDCX."
-				 " This is so far unsupported, adapt your osmo-mgw config!\n",
+			LOGPFSML(fi, LOGL_INFO, "Local MGW IuUP IP address %s changed to %s during MDCX."
+				 " Modifying RAB on HNB.\n",
 				 osmo_sockaddr_to_str(&mgw_fsm_priv->ci_hnb_crcx_ack_addr),
 				 osmo_sockaddr_to_str_buf(addr_buf, sizeof(addr_buf), &addr));
-			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+			/* Modify RAB on the HNB with the new local IuUP address (OS#6127): */
+			mgw_fsm_priv->ci_hnb_crcx_ack_addr = addr;
+			mgw_fsm_state_chg(fi, MGW_ST_ASSIGN);
 			return;
 		}
 
@@ -603,6 +645,7 @@ static const struct osmo_fsm_state mgw_fsm_states[] = {
 		.in_event_mask = S(MGW_EV_RAB_ASS_RESP),
 		.out_state_mask =
 			S(MGW_ST_MDCX_HNB) |
+			S(MGW_ST_CRCX_MSC) |
 			S(MGW_ST_FAILURE) |
 			S(MGW_ST_RELEASE),
 	},
@@ -613,6 +656,7 @@ static const struct osmo_fsm_state mgw_fsm_states[] = {
 		.in_event_mask =
 			S(MGW_EV_MGCP_OK),
 		.out_state_mask =
+			S(MGW_ST_ASSIGN) |
 			S(MGW_ST_CRCX_MSC) |
 			S(MGW_ST_FAILURE) |
 			S(MGW_ST_RELEASE),
@@ -683,6 +727,47 @@ static int release_mgw_fsm(struct hnbgw_context_map *map, struct msgb *ranap_msg
 	return 0;
 }
 
+static bool is_rab_ass_without_tli(struct hnbgw_context_map *map, ranap_message *message)
+{
+	RANAP_RAB_AssignmentRequestIEs_t *ies;
+	RANAP_ProtocolIE_ContainerPair_t *protocol_ie_container_pair;
+	RANAP_ProtocolIE_FieldPair_t *protocol_ie_field_pair;
+	RANAP_RAB_SetupOrModifyItemFirst_t _rab_setup_or_modify_item_first = { 0 };
+	RANAP_RAB_SetupOrModifyItemFirst_t *rab_setup_or_modify_item_first = &_rab_setup_or_modify_item_first;
+	int rc;
+	bool ret;
+
+	ies = &message->msg.raB_AssignmentRequestIEs;
+
+	if (!(ies->presenceMask & RAB_ASSIGNMENTREQUESTIES_RANAP_RAB_SETUPORMODIFYLIST_PRESENT))
+		return false;
+
+	/* Detect the end of the list */
+	if (ies->raB_SetupOrModifyList.list.count < 1)
+		return false;
+
+	protocol_ie_container_pair = ies->raB_SetupOrModifyList.list.array[0];
+	protocol_ie_field_pair = protocol_ie_container_pair->list.array[0];
+
+	if (!protocol_ie_field_pair)
+		return false;
+
+	if (protocol_ie_field_pair->id != RANAP_ProtocolIE_ID_id_RAB_SetupOrModifyItem) {
+		RANAP_DEBUG
+		    ("Decoding failed, the protocol IE field-pair is not of type RANAP RAB setup-or-modify-item!\n");
+		return false;
+	}
+
+	rc = ranap_decode_rab_setupormodifyitemfirst(rab_setup_or_modify_item_first,
+						     &protocol_ie_field_pair->firstValue);
+	if (rc < 0)
+		return false;
+
+	ret = rab_setup_or_modify_item_first->transportLayerInformation == NULL;
+	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_SetupOrModifyItemFirst, rab_setup_or_modify_item_first);
+	return ret;
+}
+
 /* Check if the message contains a RAB-ReleaseItem that matches the RAB-ID that is managed by the given context map */
 static bool is_our_rab_release(struct hnbgw_context_map *map, ranap_message *message)
 {
@@ -721,6 +806,15 @@ int handle_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ra
 		initialized = true;
 	}
 
+	/* There may be CS RAB Assignment Requests without actual RTP information: after a normal RAB Assignment,
+	 * another RAB Assignment may follow, modifying some RAB parameters. When there is no RTP info in the message,
+	 * there is no RTP to redirect via MGW, hence just forward as-is. */
+	if (is_rab_ass_without_tli(map, message)) {
+		LOG_MAP(map, DCN, LOGL_INFO,
+			"Rx RAB Assignment Request without transportLayerInformation, forwarding as-is\n");
+		return map_rua_dispatch(map, MAP_RUA_EV_TX_DIRECT_TRANSFER, ranap_msg);
+	}
+
 	/* The RTP stream negotiation usually begins with a RAB-AssignmentRequest and ends with an IU-Release, however
 	 * it may also be that the MSC decides to release the RAB with a dedicated RAB-AssignmentRequest that contains
 	 * a ReleaseList. In this case an FSM will already be present. */
@@ -791,6 +885,13 @@ int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *rana
 		return -1;
 	}
 
+	if (map->mgw_fi->state == MGW_ST_ESTABLISHED) {
+		/* This is a response to a second RAB Assignment Request, which only modified some RAB config on top of
+		 * an earlier RAB Assignment. Just forward the response as-is, we already have our MGW set up and need
+		 * no info from it. (i.e. we don't support modifying the RTP address.) */
+		return map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, ranap_msg);
+	}
+
 	mgw_fsm_priv = map->mgw_fi->priv;
 
 	talloc_steal(mgw_fsm_priv, ranap_msg);
